// #region History

/**
 * db-code-browser.js

 * 2020-11-08 3.9
 */

// #endregion

(function () {
    'use strict';

    // #region Constants

    const tagNames = {
        TEXTAREA: 'TEXTAREA'
    }

    const eventNames = {
        CHANGE: 'change',
        MOUSEUP: 'mouseup',
    }

    const classNames = {
        DB_ATTACHED: 'db-attached'
    }

    const attributes = {
        HREF: 'href',
        DATA_API_ROOT: 'data-api-root',
        DATA_DEFINITION_CONTROL: 'data-definition-control',
        DATA_DEFINITION_LINK_CONTROL: 'data-definition-link-control',
    }

    const selectors = {
        DB_OBJECT_LIST: 'select.db-object-list',
        DB_PERSISTENT_PROPERTIES: '.db-persistent-properties',
    }

    const formatFields = {
        URL: 'url',
        WIDTH: 'width',
        HEIGHT: 'height',
    }

    const methods = {
        GET: 'GET'
    }

    // #endregion

    // #region Global variables

    var globals = new Globals();

    /**
     * 3.9 Globals.
     *
     * @class Globals
     * @constructor
     */
    function Globals() {

        this.db = new IndexedDB('ODataDB', 'Formats', 'url');   // Local database

        this.persistentProperties = {};                         // Persistent properties

    }

    // #endregion

    // #region PageLoader

    var pageLoader = new PageLoader();

    /**
     * PageLoader.
     *
     * @class PageLoader
     * @constructor
     */
    function PageLoader() {

        this.initialLoad = initialLoad;

        /**
         * Initializes the form.
         */
        function initialLoad() {

            // Restore height and width of the controls (browser code)

            persistentProperties.restorePersistentProperties();

            configureObjectLists();

        }

        /**
         * Returns a relative URL with the document base.
         *
         * @param {string} url URL
         * @returns {string}
         */
        function addDocumentBaseURI(url) {

            if (url && !url.startsWith("http") && !url.startsWith("/"))
                url = document.baseURI + url;

            return url;
        }

        /**
         * Configures object lists (for example, used to browse object code).
         */
        function configureObjectLists() {

            // Find data controls by select.odatadb-objects
            // Such object lists are used to show database objects in the SELECT controls

            var controls = document.querySelectorAll(selectors.DB_OBJECT_LIST);

            if (controls.length == 0) {
                return false;
            }

            // Bind object lists

            for (var i = 0, count = controls.length; i < count; i++) {
                configureObjectList(controls[i]);
            }

            return true;
        }

        /**
         * Configures an object list.
         *
         * @param {Element} control Object list control.
         */
        function configureObjectList(control) {

            // 3.9 If the control is already processed, skip it

            if (control.classList.contains(classNames.DB_ATTACHED))
                return;

            // Get the URL, definition control id, and definition control itself.
            // Exit if any value is not specified

            var url = addDocumentBaseURI(control.getAttribute(attributes.DATA_API_ROOT));

            if (!url)
                return;

            var definitionControlId = control.getAttribute(attributes.DATA_DEFINITION_CONTROL);
            var definitionLinkControlId = control.getAttribute(attributes.DATA_DEFINITION_LINK_CONTROL);

            var definitionControl = definitionControlId ? document.getElementById(definitionControlId) : null;
            var definitionLinkControl = definitionLinkControlId ? document.getElementById(definitionLinkControlId) : null;

            if (!definitionControl && !definitionLinkControl)
                return;

            // Add the event listener to the object control

            control.addEventListener(eventNames.CHANGE, definitionLoader.onObjectBrowserChange, false);

            // 3.9 Set the class that the control is already processed

            control.classList.add(classNames.DB_ATTACHED);
        }

    }

    // #endregion

    // #region DefinitionLoader

    var definitionLoader = new DefinitionLoader();

    /**
     * DefinitionLoader.
     *
     * @class DefinitionLoader
     * @constructor
     */
    function DefinitionLoader() {

        this.onObjectBrowserChange = onObjectBrowserChange;

        /**
         * Handles object browser changes.
         *
         * @param {Event} e Event object.
         */
        function onObjectBrowserChange(e) {

            // Get the control and its config

            var control = e.target;

            var definitionControlId = control.getAttribute(attributes.DATA_DEFINITION_CONTROL);
            var definitionLinkControlId = control.getAttribute(attributes.DATA_DEFINITION_LINK_CONTROL);

            var definitionControl = definitionControlId ? document.getElementById(definitionControlId) : null;
            var definitionLinkControl = definitionLinkControlId ? document.getElementById(definitionLinkControlId) : null;

            if (!definitionControl && !definitionLinkControl)
                return;

            var value = control.value;

            if (value == '') {
                if (definitionControl)
                    definitionControl.value = '';
                if (definitionLinkControl)
                    definitionLinkControl.setAttribute(attributes.HREF, '#');
                return;
            }

            var url = control.getAttribute(attributes.DATA_API_ROOT) + value + '/$definition';

            if (definitionLinkControl)
                definitionLinkControl.setAttribute(attributes.HREF, url);

            if (!definitionControl)
                return;

            var xhttp = new XMLHttpRequest();

            xhttp.onreadystatechange = function () {

                if (xhttp.readyState === 4) {
                    definitionControl.value = xhttp.responseText;
                }
            };

            xhttp.open(methods.GET, url, true);
            xhttp.send();

        }

    }

    // #endregion

    // #region Database

    /**
     * 3.9 LocalDB.
     *
     * @class IndexedDB
     * @constructor
     * @param {string} dbName Database name.
     * @param {string} storeName Store name.
     * @param {string} keyPath KeyPath.
     */
    function IndexedDB(dbName, storeName, keyPath) {

        this.dbName = dbName;                   // IndexedDB database name
        this.storeName = storeName;             // IndexedDB store name
        this.keyPath = keyPath;                 // ObjectStore keyPath

        this.connectDB = connectDB;
        this.saveDbValue = saveDbValue;
        this.loadDbValue = loadDbValue;
        this.deleteDbValue = deleteDbValue;

        /**
         * Logs an error.
         *
         * @param {Object} err Error.
         */
        function logerr(err) {
            console.log(err);
        }

        /**
         * Opens a database asynchronously and calls the callback function.
         *
         * @param {Function} callback Callback function.
         */
        function connectDB(storeName, callback) {

            var indexedDB = window.indexedDB;

            var request = indexedDB.open(this.dbName, 1);

            request.onerror = logerr;

            request.onsuccess = function () {
                callback(request.result, storeName);
            }

            /** @type {IndexedDB} */
            var db = this;

            request.onupgradeneeded = function (e) {

                e.currentTarget.result.createObjectStore(storeName, { keyPath: db.keyPath });

                db.connectDB(storeName, callback);
            }
        }

        /**
         * Saves a value to a database asynchronously.
         *
         * @param {Object} value Value object including a key value property.
         */
        function saveDbValue(value) {

            this.connectDB(this.storeName, function (db, storeName) {

                var request = db.transaction(storeName, 'readwrite').objectStore(storeName).put(value);

                request.onerror = logerr;

                request.onsuccess = function () {
                    return request.result;
                }
            });
        }

        /**
         * Loads a database value and calls the callback function asynchronously.
         *
         * @param {string} key Key value.
         * @param {Function} callback Callback function.
         */
        function loadDbValue(key, callback) {

            if (!key) {
                callback(-1);
                return;
            }

            this.connectDB(this.storeName, function (db, storeName) {

                var request = db.transaction(storeName, 'readonly').objectStore(storeName).get(key);

                request.onerror = logerr;

                request.onsuccess = function () {
                    callback(request.result ? request.result : -1);
                }
            });
        }

        /**
         * Deletes a value from a database asynchronously.
         *
         * @param {null|string} key Key value.
         */
        function deleteDbValue(key) {

            if (!key) return;

            this.connectDB(this.storeName, function (db, storeName) {

                var request = db.transaction(storeName, 'readwrite').objectStore(storeName).delete(key);

                request.onerror = logerr;

                request.onsuccess = function () {
                }

            });
        }

    }

    // #endregion

    // #region Persistent properties

    /**
     * PersistentControlProperties.
     *
     * @class PersistentControlProperties
     * @constructor
     */
    function PersistentControlProperties() {

        this.control = null;                // Control
        this.height = 0;                    // offsetHeight (including paddings)
        this.width = 0;                     // offsetWidth (including paddings)
    }

    var persistentProperties = new PersistentProperties();

    /**
     * PersistentProperties.
     *
     * @class PersistentProperties
     * @constructor
     */
    function PersistentProperties() {

        this.restorePersistentProperties = restorePersistentProperties;
        this.getUrlPath = getUrlPath;

        /**
         * Restores controls with persistent properties.
         */
        function restorePersistentProperties() {

            // Find data controls by div.db-control

            var controls = document.querySelectorAll(selectors.DB_PERSISTENT_PROPERTIES);

            if (controls.length == 0) {
                return false;
            }

            // Create the required HTML controls and configurations

            for (var i = 0, count = controls.length; i < count; i++) {

                var control = controls[i];

                if (!control.id)
                    continue;

                if (!control.tagName == tagNames.TEXTAREA)
                    continue;

                // 3.9 If the control is already processed, skip it

                if (control.classList.contains(classNames.DB_ATTACHED))
                    continue;

                var properties = new PersistentControlProperties();

                properties.control = control;
                properties.height = control.offsetHeight;
                properties.width = control.offsetWidth;

                globals.persistentProperties[control.id] = properties;

                // Add the event listener to the object control

                control.addEventListener(eventNames.MOUSEUP, onPersistentPropertiesMouseUp, false);

                // 3.9 Set the class that the control is already processed

                control.classList.add(classNames.DB_ATTACHED);

                // Restore the last properties

                var key = getUrlPath(document.location.href) + "#" + control.id + '?' + selectors.DB_PERSISTENT_PROPERTIES.substring(1);

                globals.db.loadDbValue(key, function (result) {

                    if (result != -1) {                                                         // loadDbValue returns -1 on errors

                        var width = result[formatFields.WIDTH];
                        var height = result[formatFields.HEIGHT];

                        if (width && height && width >= 2 && height >= 2 && width < window.innerWidth && height < window.innerHeight) {

                            control.style.height = height + 'px';
                            control.style.width = width + 'px';

                            properties.height = control.offsetHeight;
                            properties.width = control.offsetWidth;
                        }
                    }
                });
            }

            return true;
        }

        /**
         * Handles object height changes.
         *
         * @param {Event} e Event object.
         */
        function onPersistentPropertiesMouseUp(e) {

            var control = e.target;

            if (!control.id)
                return;

            var properties = globals.persistentProperties[control.id];

            if (properties.height == control.offsetHeight && properties.width == control.offsetWidth)
                return;

            properties.height = control.offsetHeight;
            properties.width = control.offsetWidth;

            var key = getUrlPath(document.location.href) + "#" + control.id + '?' + selectors.DB_PERSISTENT_PROPERTIES.substring(1);

            var format = {};

            format[formatFields.URL] = key;
            format[formatFields.WIDTH] = properties.width;
            format[formatFields.HEIGHT] = properties.height;

            globals.db.saveDbValue(format);
        }

        /**
         * Returns a URL path.
         *
         * @param {null|string} url URL.
         */
        function getUrlPath(url) {

            if (!url) return null;

            var index = url.indexOf('#');

            if (index >= 0) {
                url = url.substring(0, index);
            }

            index = url.indexOf('?');

            if (index >= 0) {
                url = url.substring(0, index);
            }

            index = url.lastIndexOf('(');

            if (index > url.lastIndexOf('/')) {
                url = url.substring(0, index);
            }

            return url;
        }

    }

    // #endregion

    pageLoader.initialLoad();

}());
